/* ====================================================================
 *  Copyright 2004 Paul Querna
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 *
 */

#include "httpd.h"
#include "http_config.h"
#include "http_protocol.h"
#include "http_connection.h"
#include "http_request.h"
#include "http_core.h"
#include "http_log.h"
#include "apr_buckets.h"
#include "apr_strings.h"
#include "apr_tables.h"
#include "apr_thread_mutex.h"

#include "google/coredumper.h"

module AP_MODULE_DECLARE_DATA coredumper_module;

static apr_thread_mutex_t* lock;

static int coredumper_handler(request_rec * r)
{
    apr_status_t rv = APR_SUCCESS;
    apr_file_t* source;
    apr_bucket_brigade* bb;
    apr_bucket* e;
    int dump;

    if (strcmp(r->handler, "download-core")) {
        return DECLINED;
    }

    r->allowed |= (AP_METHOD_BIT << M_GET);
    if (r->method_number != M_GET) {
        return DECLINED;
    }

    apr_thread_mutex_lock(lock);
    dump = GetCoreDump();
    apr_thread_mutex_unlock(lock);

    if (dump < 0) {
        /* Some kind of error. Log the errno. */
        ap_log_rerror(APLOG_MARK, APLOG_ERR, errno, r, 
            "GetCoreDump() Failed. Return Value: %d", dump);
        return HTTP_INTERNAL_SERVER_ERROR;
    }

    rv = apr_os_file_put(&source, &dump, APR_READ, r->pool);
    if (rv != APR_SUCCESS) {
        ap_log_rerror(APLOG_MARK, APLOG_ERR, rv, r, 
            "failed to convert dump into an apr_file_t");
        return HTTP_INTERNAL_SERVER_ERROR;
    }

    bb = apr_brigade_create(r->pool, r->connection->bucket_alloc);

    e = apr_bucket_pipe_create(source, r->connection->bucket_alloc);

    APR_BRIGADE_INSERT_TAIL(bb, e);

    ap_set_content_type(r, "application/x-coredump");

    return ap_pass_brigade(r->output_filters, bb);
}

static void child_init(apr_pool_t *p, server_rec *s)
{
    apr_thread_mutex_create(&lock, 0, p);
}

static void coredumper_hooks(apr_pool_t * p)
{
    ap_hook_child_init(child_init, NULL, NULL, APR_HOOK_FIRST);
    ap_hook_handler(coredumper_handler, NULL, NULL, APR_HOOK_FIRST);
}

static const command_rec coredumper_cmds[] = {
    {NULL}
};

module AP_MODULE_DECLARE_DATA coredumper_module = {
    STANDARD20_MODULE_STUFF,
    NULL,
    NULL,
    NULL,
    NULL,
    coredumper_cmds,
    coredumper_hooks
};
